import re
from collections import namedtuple

#===============================================================================
# Basic struct classes 
#===============================================================================

class BaseStruct(object):
    def __init__(self, **kwargs):
        for slt in self.__slots__:
            if slt in kwargs:
                setattr(self, slt, kwargs[slt])

    def __repr__(self):
        return (self.__class__.__name__ + "(" + ", ".join(
            [s + "=" + str(getattr(self, s)) for s in self.__slots__]) + ")")

class BaseLine(BaseStruct):
    __slots__ = ('ln', 'linetext')

class ErrorLine(BaseLine):
    __slots__ = BaseLine.__slots__ + ('error',)

class Line(BaseLine):
    __slots__ = BaseLine.__slots__ + ('label',)

class Directive(Line):
    __slots__ = Line.__slots__ + ('directive', 'params')

class Command(Line):
    __slots__ = Line.__slots__ + ('opcode', 'type', 'comb', 'opsrc', 'opdst')

class Operand(BaseStruct):
    pass

class Number(Operand):
    __slots__ = ('value',)

class Register(Operand):
    __slots__ = ('register',)

class Label(Operand):
    __slots__ = ('label',)

class Relative(Label):
    __slots__ = Label.__slots__ + ('offset',)

#===============================================================================
# Reserved names
#===============================================================================

OpcodeRule = namedtuple("OpcodeRule", ['cls', 'value', 
                                       'illegal_opsrc', 
                                       'illegal_opdst'])

OPCODES = {  # opcode string : opcode rule
                  'mov': OpcodeRule("A", 0, (), (Number,)),
                  'cmp': OpcodeRule("A", 1, (), ()),
                  'add': OpcodeRule("A", 2, (), (Number,)),
                  'sub': OpcodeRule("A", 3, (), (Number,)),
                  'lea': OpcodeRule("A", 6, (Number,), (Number,)),
                  'not': OpcodeRule("B", 4, (), (Number,)),
                  'clr': OpcodeRule("B", 5, (), (Number,)),
                  'inc': OpcodeRule("B", 7, (), (Number,)),
                  'dec': OpcodeRule("B", 8, (), (Number,)),
                  'jmp': OpcodeRule("B", 9, (), (Number,)),
                  'bne': OpcodeRule("B", 10, (), (Number,)),
                  'red': OpcodeRule("B", 11, (), (Number,)),
                  'prn': OpcodeRule("B", 12, (), ()),
                  'jsr': OpcodeRule("B", 13, (), (Number, Register, Relative)),
                  'rts': OpcodeRule("C", 14, (), ()),
                  'stop': OpcodeRule("C", 15, (), ()),
                  }

CLASSA_OPCODES = [op for op, rule in OPCODES.iteritems() if rule.cls == "A"]
CLASSB_OPCODES = [op for op, rule in OPCODES.iteritems() if rule.cls == "B"]
CLASSC_OPCODES = [op for op, rule in OPCODES.iteritems() if rule.cls == "C"]

DIRECTIVES = ['data', 'string', 'extern', 'entry']

GP_REGISTERS = ['r0', 'r1', 'r2', 'r3', 'r4', 'r5', 'r6', 'r7']
SP_REGISTERS = ['PC', 'SP', 'PSW']

# only General Purpose registers can be referenced
# so special registers won't be considered reserved words

RESERVED_WORDS = OPCODES.keys() + GP_REGISTERS

#===============================================================================
# Regular expressions 
#===============================================================================

_number = r"[+-]?\d+"
_label = "[a-zA-Z][a-zA-Z0-9]{,30}"
_register = "|".join(GP_REGISTERS)
_classa = "|".join(CLASSA_OPCODES)
_classb = "|".join(CLASSB_OPCODES)
_classc = "|".join(CLASSC_OPCODES)
_directives = "|".join(DIRECTIVES)

empty = re.compile(r"^\s*$")
comment = re.compile(r"^;")
labelled_line = re.compile(r"^(?P<LABEL>%s):(?P<REST>.*)$" % _label) 
unlabelled_directive = re.compile(r"^\.(?P<DIRECTIVE>%s)(?P<PARAMS>\s.*)?$" 
                                  % _directives)
data_params = re.compile(r"^(%s)(\s*,\s*%s)*$" % (_number, _number))
string_param = re.compile(r"^\"(?P<STRING>.*)\"$")

_classab = r"^(?P<OPCODE>%s)/(?P<TYPE>0|1/[01]/[01])\s(?P<OPERANDS>.*)$"
classa_command = re.compile(_classab % _classa)
classb_command = re.compile(_classab % _classb)
classc_command = re.compile(r"^(?P<OPCODE>%s)/0$" % _classc)

classa_operands = re.compile(r"^(?P<OPSRC>[^,]*),(?P<OPDST>[^,]*)$") 
classb_operands = re.compile(r"^(?P<OPDST>[^,]*)$")

immediate_operand = re.compile(r"^#(?P<NUMBER>%s)$" % _number)
register_operand = re.compile(r"^(?P<REGISTER>%s)$" % _register)
label_operand = re.compile(r"^(?P<LABEL>%s)$" % _label)
mixed_operand = re.compile(r"^(?P<BASELABEL>%s)\{(?P<OFFSET>.*)\}$" % _label)

number_value = re.compile(r"^(?P<NUMBER>%s)$" % _number)
label_value = label_operand
register_value = register_operand

#===============================================================================
# Parsers
#===============================================================================

def parse_opcode_type(match_obj):
    """ Parse an opcode TYPE and COMB values.
    """
    opcode = match_obj.group('OPCODE')
    type_comb = match_obj.group('TYPE')
    optype = comb = 0
    if type_comb[0] == "1":
        optype = 1
        comb = int(type_comb[2] + type_comb[4], 2)
    return opcode, optype, comb 
    
def parse_operand(operand_str):
    """ Parse an operand of a command.
    """
    operand_str = operand_str.strip()
    
    check = immediate_operand.match(operand_str)
    if check:
        return Number(value=int(check.group('NUMBER')))
    
    check = register_operand.match(operand_str)
    if check:
        return Register(register=check.group('REGISTER'))
    
    # must be after register check
    check = label_operand.match(operand_str)
    if check:
        return Label(label=check.group('LABEL')) 

    check = mixed_operand.match(operand_str)
    if check:
        baselabel = check.group('BASELABEL')
        offset = check.group('OFFSET').strip()
        check = number_value.match(offset)
        if check:
            return Relative(label=baselabel, 
                offset=Number(value=int(check.group('NUMBER'))))
        check = register_value.match(offset)
        if check:
            return Relative(label=baselabel, 
                offset=Register(register=check.group('REGISTER')))
        # must be after label check
        check = label_value.match(offset)
        if check:
            return Relative(label=baselabel, 
                offset=Label(label=check.group('LABEL')))
            
    return None

def is_valid_addressing(opcode, opsrc=None, opdst=None):
    """ Test if addressing combination of operands is allowed for the opcode.
    """
    return not (isinstance(opsrc, OPCODES[opcode].illegal_opsrc) or 
                isinstance(opdst, OPCODES[opcode].illegal_opdst))
    
def parse_file(fname):
    """ Parse a file into instruction objects.
    
    Return a list of objects describing the meaningful lines in file.
    Ignores blank and comment lines. Erroneous lines result in error objects.
    """
    ops = []
    
    with open(fname, "r") as fin:
        for ln, line in enumerate(fin, 1):
            linetext = line.strip("\n")
            errortext = "Invalid instruction"
            newop = ErrorLine(ln=ln, linetext=linetext, error=errortext)                
            
            # empty or comment lines
            if empty.match(line) or comment.match(line):
                continue
            
            # check for label
            label = ''
            check = labelled_line.match(line)
            if check:
                label = check.group('LABEL')
                line = check.group('REST')
            line = line.strip()
            
            # directives
            check = unlabelled_directive.match(line)
            if check:
                directive = check.group('DIRECTIVE')
                params = check.group('PARAMS')
                params = params.strip() if params else params
                
                if not params:
                    # disallow directives without parameters
                    newop.error = "No parameters provided for directive"
                else:
                    if directive == 'data' and data_params.match(params):
                        newop = Directive(ln=ln, linetext=linetext, label=label, 
                                  directive=directive,
                                  params=[int(i) for i in params.split(',')])
                    elif directive == 'string':
                        check = string_param.match(params)
                        if check:
                            newop = Directive(ln=ln, linetext=linetext, 
                                    label=label, directive=directive,
                                    params=check.group('STRING'))
                    elif directive == 'extern' or directive == 'entry':
                        check = label_value.match(params)
                        if check:
                            # .entry and .extern labels are meaningless
                            newop = Directive(ln=ln, linetext=linetext, 
                                    label="", directive=directive,
                                    params=check.group(0))
                    else:
                        newop.error = "Invalid directive parameter(s)"
                        
                ops.append(newop)
                continue

            # two operand commands            
            check = classa_command.match(line)
            if check:
                opcode, optype, comb = parse_opcode_type(check)
                operands = check.group('OPERANDS').strip()
                
                check = classa_operands.match(operands)
                if check:
                    opsrc = parse_operand(check.group('OPSRC'))
                    opdst = parse_operand(check.group('OPDST'))
                    if opsrc and opdst:
                        if is_valid_addressing(opcode, opsrc, opdst):
                            newop = Command(ln=ln, linetext=linetext, 
                                    label=label, opcode=opcode, type=optype, 
                                    comb=comb, opsrc=opsrc, opdst=opdst)
                        else:
                            newop.error = \
                                    "Invalid operand type(s) for this opcode"
                    else:
                        newop.error = "Invalid operand(s)"
                else:
                    newop.error = "Invalid operand(s)"
                        
                ops.append(newop)
                continue
                
            # one operand commands            
            check = classb_command.match(line)
            if check:
                opcode, optype, comb = parse_opcode_type(check)
                operands = check.group('OPERANDS').strip()
                
                check = classb_operands.match(operands)
                if check:
                    opdst = parse_operand(check.group('OPDST'))
                    if opdst:
                        if is_valid_addressing(opcode, opdst=opdst):
                            newop = Command(ln=ln, linetext=linetext, 
                                    label=label, opcode=opcode, type=optype, 
                                    comb=comb, opsrc=None, opdst=opdst)
                        else:
                            newop.error = \
                                    "Invalid operand type(s) for this opcode"
                    else:
                        newop.error = "Invalid operand(s)"
                else:
                    newop.error = "Invalid operand(s)"
                        
                ops.append(newop)
                continue
                
            # no operand commands            
            check = classc_command.match(line)
            if check:
                opcode = check.group('OPCODE')
                optype = comb = 0
                newop = Command(ln=ln, linetext=linetext, label=label, 
                                 opcode=opcode, type=optype, comb=comb,
                                 opsrc=None, opdst=None)
                        
                ops.append(newop)
                continue

            ops.append(newop)
            
    return ops
    
def map_and_validate_labels(ops):
    """ Validate the labels in ops list.
    
    Find duplicate labels, missing labels and replace the relevant lines
    in ops with errors.
    Return a list of entry labels.
    """
    
    # map label definitions, check for duplicates
    def_labels = {} # {label name: index in ops} 
    for idx, op in enumerate(ops):
        if isinstance(op, (Directive, Command)) and op.label:
            if op.label in def_labels:
                ops[idx] = ErrorLine(ln=op.ln, linetext=op.linetext, 
                                    error="Duplicate label definition")
            else:
                def_labels[op.label] = idx

    # map entry labels, check for missing labels
    ent_labels = [] 
    for idx, op in enumerate(ops):
        if isinstance(op, Directive) and op.directive == 'entry' and op.params:
            if op.params not in def_labels:
                ops[idx] = ErrorLine(ln=op.ln, linetext=op.linetext, 
                                error="No matching label for entry directive")
            else:
                ent_labels.append(op.params)
            
    # map extern labels, check for label collisions
    ext_labels = {} # {index in ops: label name}
    for idx, op in enumerate(ops):
        if isinstance(op, Directive) and op.directive == 'extern' and op.params:
            if op.params in def_labels:
                ops[idx] = ErrorLine(ln=op.ln, linetext=op.linetext, 
                        error="Local label can't be used for extern directive")
            else:
                ext_labels[idx] = op.params
            
    # check all other label references
    for idx, op in enumerate(ops):
        if not isinstance(op, Command):
            continue
        line_labels = []
        if isinstance(op.opsrc, Label):
            line_labels.append(op.opsrc.label)        
        if isinstance(op.opdst, Label):
            line_labels.append(op.opdst.label)        
        if isinstance(op.opsrc, Relative) and isinstance(op.opsrc.offset, Label):
            line_labels.append(op.opsrc.offset.label)        
        if isinstance(op.opdst, Relative) and isinstance(op.opdst.offset, Label):
            line_labels.append(op.opdst.offset.label)        

        bad_labels = [lbl for lbl in line_labels 
                      if lbl not in (def_labels.keys() + ext_labels.values())]
        if bad_labels:
            ops[idx] = ErrorLine(ln=op.ln, linetext=op.linetext, 
                error="Undefined label(s) " + ", ".join(bad_labels))
    
    return ent_labels
    
    
