from consts import (WORD_SIZE_QUBITS, WORD_SIZE_BITS, QUBITMASK, QUBITS, QUBASE)

def int2bits(val, exact_width=WORD_SIZE_BITS):
    """ Convert integer into string representing bits, fitted to exact width.
    
    If `val` is negative, convert according to 2's complement.
    If exact width is shorted than needed for representation, representation is
    truncated.
    """
    return bin(val & ((1 << exact_width) - 1)).lstrip('-0b').zfill(exact_width)

def bits2qubits(bs, output_width=WORD_SIZE_QUBITS):
    """ Convert a string representing bits into string representing qubits.
    """
    bint = int(bs, 2)
    qstr = []
    while bint:
        qstr = [hex(bint & QUBITMASK)[-1]] + qstr
        bint >>= QUBITS
    return "".join(qstr).zfill(output_width)

def int2qubits(val, bit_exact_width=WORD_SIZE_BITS,
               qubit_fill_width=WORD_SIZE_QUBITS):
    """ Convert integer into string representing qubits (via bitstr conversion).
    
    `bit_exact_width` is required for the exact width of the itermediate
    bit string (esp. important for negative values).
    `qubit_fill_width` is the width to which the output is 
    padded.
    """
    return bits2qubits(int2bits(val, bit_exact_width), qubit_fill_width)
    

def qubits2bits(qs, output_width=WORD_SIZE_BITS):
    """ Convert a string representing qubits into string representing bits.
    """
    qint = int(qs, QUBASE)
    return bin(qint)[2:].zfill(output_width)

class BitField(object):
    def __init__(self, size, init_val=0):
        self._size = size
        self._value = init_val
        
    def __len__(self):
        return self._size
    
    def set(self, value):
        self._value = value
        
    def get(self):
        return self._value
    
    def encode(self):
        return int2bits(self._value, self._size)
    
    def __str__(self):
        return str(self.get())

class BitStrStruct(object):
    # in order, MSB first
    __fields__ = []
    __size_hints__ = []
    
    def __init__(self):
        for name, size in zip(self.__fields__, self.__size_hints__):
            setattr(self, name, BitField(size))
            
    def __len__(self):
        return sum(self.__size_hints__)
        
    def encode_bits(self):
        res = [getattr(self, name).encode() for name in self.__fields__]
        return "".join(res)
        
    def __repr__(self):
        return self.encode_bits()
        
class CommandStr(BitStrStruct):
    __fields__ = ['unused', 'type', 'opcode', 'src_addr', 'src_reg',
                 'dst_addr', 'dst_reg', 'comb']
    __size_hints__ = [3, 1, 4, 2, 3, 2, 3, 2]
    
