#!/usr/bin/env python
import os
import sys

from consts import *
from bithandler import *
from asparse import *

def addressing_value(operand):
    """ Return int values of addressing and register fields for `operand`.
    """
    if isinstance(operand, Relative):
        if isinstance(operand.offset, Register):
            return MIXED_ADRS, int(operand.offset.register[1:])
        return MIXED_ADRS, 0
    if isinstance(operand, Label):
        return DIRECT_ADRS, 0
    if isinstance(operand, Register):
        return REGISTER_ADRS, int(operand.register[1:])
    if isinstance(operand, Number):
        return IMMEDIATE_ADRS, 0
    return 0, 0

def extended_params(operand):
    """ Return the additional words needed for an operand.
    
    Return either one or two element list of (word, A.R.E) tuples.
    Label names are returned as string text (with IS_RELOC A.R.E. value). 
    """
    ret = []
    if isinstance(operand, Relative):
        ret.append((operand.label, IS_RELOC))
        if isinstance(operand.offset, Number):
            ret.append((int2bits(operand.offset.value), IS_ABS))
        elif isinstance(operand.offset, Label):
            ret.append((operand.offset.label, IS_RELOC))
    elif isinstance(operand, Label):
        ret.append((operand.label, IS_RELOC))
    elif isinstance(operand, Number):
        ret.append((int2bits(operand.value), IS_ABS))
    return ret
    

def first_pass(ops):
    """ Perform first assembly pass, not resolving labels yet.
    
    Return:
    - text section: list of (bitword, A.R.E. value) tuples
    - text labels: dict of label location relative to beginning of text
    - data section: list of bitwords
    - data_labels: dict of label location relative to beginning of data
    """
    text = []
    text_labels = {}
    data = []
    data_labels = {}
    
    for op in ops:
        if isinstance(op, Directive) and op.directive == 'data':
            if op.label:
                data_labels[op.label] = len(data)
            data.extend([int2bits(v) for v in op.params])
        elif isinstance(op, Directive) and op.directive == 'string':
            if op.label:
                data_labels[op.label] = len(data)
            data.extend([int2bits(ord(c)) for c in (op.params + "\0")])
        elif isinstance(op, Command):
            if op.label:
                text_labels[op.label] = len(text)
            cmd = CommandStr()
            cmd.opcode.set(OPCODES[op.opcode].value)
            cmd.type.set(op.type)
            cmd.comb.set(op.comb)
            ad, rg = addressing_value(op.opsrc)
            cmd.src_addr.set(ad)
            cmd.src_reg.set(rg) 
            ad, rg = addressing_value(op.opdst)
            cmd.dst_addr.set(ad)
            cmd.dst_reg.set(rg)
            text.append((cmd.encode_bits(), IS_ABS))
            text.extend(extended_params(op.opsrc))
            text.extend(extended_params(op.opdst))
    
    return text, text_labels, data, data_labels

def second_pass(text, text_labels, data, data_labels):
    """ Perform second assembly pass.
    
    Assign memory addresses, resolve local and external labels.
    Returns:
    - first line with text size and data size in qubits
    - a list of strings representing machine code
    - a list of strings representing externals file
    """
    text_size = len(text)
    data_size = len(data)
    code = []
    ext = []
    addr = START_ADDR
    for instr, are in text:
        qubit_addr = int2qubits(addr, qubit_fill_width=0)
        if are == IS_ABS:
            code.append(" ".join([qubit_addr, bits2qubits(instr), are]))
        elif are == IS_RELOC:
            if instr in text_labels:
                lbl_addr = START_ADDR + text_labels[instr]
            elif instr in data_labels:
                lbl_addr = START_ADDR + text_size + data_labels[instr]
            else:
                lbl_addr = 0
                are = IS_EXT
                ext.append(" ".join([instr, qubit_addr]))
            code.append(" ".join([qubit_addr, int2qubits(lbl_addr), are]))
        addr += 1

    for data_val in data:
        qubit_addr = int2qubits(addr, qubit_fill_width=0)
        code.append(" ".join([qubit_addr, bits2qubits(data_val)]))
        addr += 1
    
    return " ".join([int2qubits(text_size, qubit_fill_width=1), 
                     int2qubits(data_size, qubit_fill_width=1)]), code, ext

def error_pass(ops):
    """ Print errors if any exist. """
    has_errors = False
    for op in ops:
        if isinstance(op, ErrorLine):
            has_errors = True
            print "Error in line %d: %s" % (op.ln, op.error)
    return has_errors

def make_entries(entry_labels, len_text, text_labels, data_labels):
    """ Return the entries for the ent output file. """
    # text_labels first, to comply with example
    ent = [lbl + " " + int2qubits(START_ADDR + text_labels[lbl], 
                                  qubit_fill_width=0)
            for lbl in entry_labels if lbl in text_labels]
    
    ent.extend([lbl + " " + int2qubits(START_ADDR + len_text + data_labels[lbl], 
                                       qubit_fill_width=0)
                for lbl in entry_labels if lbl in data_labels])
    return ent
    
def get_args():
    """ Get input argument, exit on error, return necessary filenames.
    """
    if (len(sys.argv) != 2 or sys.argv[1] == AS_SUFFIX or  
        not sys.argv[1].endswith(AS_SUFFIX)):
        print "usage: mainasm <*.as source filename>"
        raise SystemExit(1)
    
    if not os.path.isfile(sys.argv[1]):
        print "Error: source file %s not found." % sys.argv[1]
        raise SystemExit(1)
        
    fas = sys.argv[1]
    fob = fas[:-len(AS_SUFFIX)] + OB_SUFFIX
    fent = fas[:-len(AS_SUFFIX)] + ENT_SUFFIX
    fext = fas[:-len(AS_SUFFIX)] + EXT_SUFFIX
    
    return fas, fob, fent, fext
    
def filewrite(fname, data):
    """ If `data` string list not empty, write to `fname` file with linebreaks.
    """
    if data:
        with open(fname, "w") as f:
            for line in data:
                f.write(line + "\n")
    
def main():
    fas, fob, fent, fext = get_args()
    
    ops = parse_file(fas)
    entry_labels = map_and_validate_labels(ops)
    if error_pass(ops):
        raise SystemExit(1)
    text, text_labels, data, data_labels = first_pass(ops)
    sizes, code, extfile = second_pass(text, text_labels, data, data_labels)
    entfile = make_entries(entry_labels, len(text), text_labels, data_labels)
    
    filewrite(fob, [sizes] + code)
    filewrite(fext, extfile)
    filewrite(fent, entfile)

if __name__ == "__main__":
    main()